/* 
  Magnum 
  Copyright (C) 2002-2005 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics.h"
#include "gui.h"
#include "sky.h"

#include "graphics.inl"
#include "gui.inl"


namespace magnum { namespace sample {
using namespace magnum::graphics;
using namespace magnum::graphics3d;
using namespace magnum::physics;
using namespace magnum::object;
using namespace magnum::io;
using namespace magnum::math;
using magnum::system::Timer;
using magnum::stats::Normal;
using magnum::debug::Logger;

#define TimeStep				5

typedef TrigonometricPolynomial<float, Vector3<float> >	BorderFunction;

static Camera*				g_Camera;
static Mode*				g_DefaultMode;
static Mode*				g_TransparentMode;
static SpringModel*			g_Model;
static SkyModel*			g_Sky;
static PointLight*			g_Light;
static ParameterBuffer*		g_Parameters;
static Relaxator<float,3>*	g_Springs;
static BorderFunction		g_Border1, g_Border2, g_BlendedBorder;
static Timer				g_Timer;
static Normal				g_Random;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class F> void applyFunction(F& func) {
	unsigned int i;
	for (i = 0; i < 50; i++)
		(*g_Springs)[i].setPosition(func(i/196.0f));
	for (i = 1; i < 49; i++)
		(*g_Springs)[i*50+49].setPosition(func((i+50)/196.0f));
	for (i = 0; i < 50; i++)
		(*g_Springs)[49*50+49-i].setPosition(func((i+98)/196.0f));
	for (i = 1; i < 49; i++)
		(*g_Springs)[(49-i)*50].setPosition(func((i+148)/196.0f));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline Vector3<float> evalBorder(float v) {
	return g_BlendedBorder(v * 2* (float)M_PI);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static void initBorders() {
	unsigned int i;
	
	g_Random.setVariance(16000.0f);
	g_Border1.setDegree(3);
	g_Border1[0].x = g_Border1[0].y = g_Border1[0].z = 0.0f;
	for (i = 1; i <= g_Border1.getDegree()*2; ++i) {
		g_Border1[i].x = g_Random();
		g_Border1[i].y = g_Random();
		g_Border1[i].z = g_Random();
	}
	
	g_Border2.setDegree(3);
	g_Border2[0].x = g_Border2[0].y = g_Border2[0].z = 0.0f;
	for (i = 1; i <= g_Border2.getDegree()*2; ++i) {
		g_Border2[i].x = g_Random();
		g_Border2[i].y = g_Random();
		g_Border2[i].z = g_Random();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static void nextBorders() {
	g_Border1 = g_Border2;
	for (unsigned int i = 1; i <= g_Border2.getDegree()*2; ++i) {
		g_Border2[i].x = g_Random();
		g_Border2[i].y = g_Random();
		g_Border2[i].z = g_Random();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onInit() {
	// Create Camera
	g_Camera = new Camera();
	g_Camera->setNearPlane(10);
	g_Camera->setFarPlane(1700);
	g_Camera->select();

	// Create Render Modes
	g_DefaultMode = new Mode();
	g_DefaultMode->setDepthTest(Mode::Less);
	g_DefaultMode->setDepthWrite(true);
	
	g_TransparentMode = new Mode();
	g_TransparentMode->setDepthTest(Mode::Less);
	g_TransparentMode->setDepthWrite(false);
	g_TransparentMode->setBlendMode(Mode::SrcColour, Mode::DstColour);
	g_TransparentMode->setBlendEquation(Mode::Add);
//	g_TransparentMode->setBlendColour(Colour<float>(0.3f,0.3f,0.3f));
	
	// Create Light
	g_Light = new PointLight();
	g_Light->setPosition(Vector3<float>(900.0f,790.0f,400.0f));
	g_Light->setAmbient(Colour<float>(0.2f,0.2f,0.2f));
	g_Light->setDiffuse(Colour<float>(1.0f,1.0f,1.0f));
	g_Light->setSpecular(Colour<float>(0.3f,0.3f,0.3f));
	
	// Create Skybox
	g_Sky = createSky();
	
	// Create other stuff
	AutoPointer<File> texf = SystemFile::open("data\\soapfilm.jpg");
	AutoPointer<Image> img = TranslatorJPG().decode(texf->getInputStream());
	AutoPointer<Texture> texture = Texture2D::create(img, Texture::GenerateAll);
	texture->setFilter(Texture::LinearMipmapLinear, Texture::Linear);

	AutoPointer<ColourMaterial> material = new ColourMaterial();
	material->setAmbient(Colour<float>(0.2f, 0.1f, 0.1f));
	material->setDiffuse(Colour<float>(1.0f, 0.8f, 0.9f));
	material->setSpecular(Colour<float>(1.0f, 0.8f, 0.7f));
	material->setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	
	AutoPointer<File> vertf = SystemFile::open("envshader.vp");
	AutoPointer<VertexProgram> vprog = VertexProgram::create(Content(vertf));
	
	AutoPointer<File> pixf = SystemFile::open("envshader.pp");
	AutoPointer<PixelProgram> pprog = PixelProgram::create(Content(pixf));

	AutoPointer<ProgrammableShader> shader = ProgrammableShader::create();
	shader->attach(vprog);
	shader->attach(pprog);
	shader->link();

	AutoPointer<Appearance> appearance = new Appearance();
	appearance->attachShader(shader);
	appearance->attachMaterial(material);
	appearance->attachTexture(texture,0);
	appearance->attachTexture(g_Sky->getAppearance()->getTexture(0),1);

	g_Model = new SpringModel(SpringModel::Rectangle,50,5.0f);
	g_Model->attachAppearance(appearance);
	g_Springs = g_Model->getSprings();
	for (unsigned int i = 0; i < 50; i++) {
		(*g_Springs)[i].setFixed(true);
		(*g_Springs)[i*50].setFixed(true);
		(*g_Springs)[i*50+49].setFixed(true);
		(*g_Springs)[49*50+i].setFixed(true);
	}
	g_Springs->setGravity(Vector3<float>(0.0f, 0.0f, 0.0f));
	g_Springs->setConstant(60.0f);
	g_Springs->setDamping(0.2f);

	g_Parameters = ParameterBuffer::create(g_Model->getParameters());
	
	// Create border functions
	initBorders();
	g_BlendedBorder = g_Border1;
	applyFunction(evalBorder);
	for (i = 0; i < 1000; i++)
		g_Model->relax(1.0f/40.0f);

	g_Timer.init(0);
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onExit() {
	g_Model->release();
	g_Sky->release();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool GraphicsView::onHandleMessage(const Message& msg) {
	return OpenGL::onHandleMessage(msg);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicsView::onRender() {
	g_DefaultMode->select();
	m_Frame->clear(FrameBuffer::ClearColour | FrameBuffer::ClearDepth | FrameBuffer::ClearStencil);

	g_Sky->render(g_Parameters);

	g_TransparentMode->select();
	m_MatrixStack->selectMode(MatrixStack::ModelView);
	m_MatrixStack->identity();

	g_Light->select(0);

	m_Context->setCulling(Context::CullNone);
	*(Matrix4<float>*)g_Parameters->getBuffer() = (Matrix4<float>)getTransform();
	g_Model->update();
	g_Model->render(g_Parameters);

	g_Timer.update();
//	float t = min(5.0f, (float)((double)g_Timer.getCurrent() / (double)Timer::Second));
	float t = (float)((double)g_Timer.getCurrent() / (double)Timer::Second);
	if (t > TimeStep) {
		nextBorders();
		g_Timer.init(g_Timer.getCurrent() - TimeStep*Timer::Second);
		t = (float)((double)g_Timer.getCurrent() / (double)Timer::Second);
	}
	g_BlendedBorder = t/TimeStep*g_Border2 + (TimeStep-t)/TimeStep*g_Border1;
	applyFunction(evalBorder);

	g_Model->relax(1.0f/40.0f);
	g_Model->relax(1.0f/40.0f);
}

}; };
